# -*- coding: utf-8 -*-
################################################################################
#
#  pyScanMaster -- Python Interface to ScanMaster POS
#  Copyright © 2013-2014 Sacramento Natural Foods Co-op, Inc
#
#  This file is part of pyScanMaster.
#
#  pyScanMaster is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  pyScanMaster is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  pyScanMaster.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
ASCII Data
"""

import re
from struct import unpack


# The one and only "blank" generic field.
BLANK = object()

record_length_pattern = re.compile(r'^(?P<length>\d+)\,(?P<data>.*)$', re.DOTALL)


class Record(object):
    """
    Generic class, to abstract the concept of a data record, as found within an
    ASCII file which has been exported by the ``BUTIL.EXE`` Pervasive utility.
    """

    def __init__(self, data_tuple=None):
        if data_tuple is not None:
            assert len(data_tuple) == len(self.data_format)
            self.data_tuple = data_tuple
        else:
            self.data_tuple = None

        self.field_map = {}
        for i, (start, length, fieldname) in enumerate(self.data_format):
            if fieldname is not BLANK:
                self.field_map[fieldname] = i

    def __getattr__(self, name):
        if name in self.field_map:
            assert self.data_tuple is not None
            return self.data_tuple[self.field_map[name]]

    @classmethod
    def struct_format(cls):
        """
        The binary data format string, as used by ``struct``.  This format is
        automatically generated from the contents of the :attr:`data_format`
        attribute which must be defined on each record class.
        """
        lengths = ['{0}s'.format(f[1]) for f in cls.data_format]
        return '<{0}'.format(''.join(lengths))


class RecordParser(object):
    """
    Generic parser for ASCII data records.
    """

    def __init__(self, record_class):
        self.record_class = record_class

    def parse_records(self, data_path):
        struct_format = self.record_class.struct_format()
        records = []
        with open(data_path, 'rb') as data_file:
            data = ''
            for line in data_file:
                data += line
                if data.endswith('\r\n'):
                    data = data[:-2]
                    match = record_length_pattern.match(data)
                    if match:
                        assert len(match.group('data')) == int(match.group('length'))
                        data = match.group('data')
                    records.append(self.parse_record(struct_format, data))
                    data = ''
        return records

    def parse_record(self, struct_format, data):
        return self.record_class(unpack(struct_format, data))
