# -*- coding: utf-8 -*-
################################################################################
#
#  pyScanMaster -- Python Interface to ScanMaster POS
#  Copyright © 2013-2015 Sacramento Natural Foods Co-op, Inc
#
#  This file is part of pyScanMaster.
#
#  pyScanMaster is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  pyScanMaster is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  pyScanMaster.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Models
"""

from __future__ import unicode_literals

import logging

import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, object_mapper
from sqlalchemy.ext.associationproxy import association_proxy

import datetime
from decimal import Decimal


__all__ = [
    'Department', 'Item', 'AdjectivePrice', 'ElectronicCoupon',
    'Customer', 'NegativeCheck', 'InstoreCharge', 'ChargeHistory',
    'EJTransaction', 'EJCustomer', 'EJItem', 'EJPromotion', 'EJDiscount', 'EJTax',
    ]

log = logging.getLogger(__name__)


class BaseObject(object):
    """
    Superclass for the declarative base.
    """

    def __init__(self, **kwargs):
        """
        Custom constructor to provide default attribute values.

        This leverages column defaults in order to initially populate the
        object instance.
        """
        for attr in object_mapper(self).column_attrs:
            if attr.key not in kwargs:
                column = attr.columns[0]
                if column.default:
                    setattr(self, attr.key, column.default.arg)
        # Kwargs always overrides default values.  This is especially necessary
        # since some attributes are "raw" columns and others use property
        # function wrappers.  Therefore the property will be set after the
        # column has provided the default value.
        for key, value in kwargs.iteritems():
            setattr(self, key, value)
    

Base = declarative_base(cls=BaseObject, constructor=None)


# NOTE: The following comments and code appeared to be necessary at some point,
# but I have since been able to access a ScanMaster database just fine on Linux
# without it.  I used the Pervasive `TranslationOption` setting which may have
# made the difference (?).

# # There seems to be some sort of Unicode issue with (at least) the Linux /
# # Python 2.7 / pyodbc / Pervasive combination.  Not sure "whose fault" it is
# # exactly but I was seeing some nasty premature application death.  The only
# # thing I found to resolve the issue was to force SQLAchemy to convert to/from
# # Unicode and keep things "safe" on the database side.
# def String(*args, **kwargs):
#     kwargs.setdefault('convert_unicode', 'force')
#     return SA_String(*args, **kwargs)


def date_property(column, fmt):
    """
    Establish a date property for the given column.
    """
    
    def getter(self):
        value = getattr(self, column)
        if value is None:
            return None
        if value == '000000':
            return None
        return datetime.datetime.strptime(value, fmt).date()

    def setter(self, value):
        if value is None:
            setattr(self, column, '000000')
        elif isinstance(value, basestring):
            setattr(self, column, value)
        else:
            setattr(self, column, value.strftime(fmt))

    return property(getter, setter)


def decimal_property(column, length):
    """
    Establish a decimal property for the given column.
    """

    def getter(self):
        value = getattr(self, column)
        if value is None:
            return None
        return Decimal('{0}.{1}'.format(value[:-2], value[-2:]))

    def setter(self, value):
        if value is None:
            setattr(self, column, None)
        elif isinstance(value, basestring):
            setattr(self, column, value)
        else:
            fmt = '{{0:0{0}d}}'.format(length)
            setattr(self, column, fmt.format(int(value * 100)))

    return property(getter, setter)


def int_property(column, length):
    """
    Establish an integer property for the given column.
    """

    def getter(self):
        value = getattr(self, column)
        if value is None:
            return 0
        try:
            return int(value)
        except ValueError:
            log.warning("invalid integer value {0} for column {1}: {2}".format(
                    repr(value), repr(column), repr(self)))
            return 0

    def setter(self, value):
        if value is None:
            value = 0
        else:
            value = int(value)
        fmt = '{{0:0{0}d}}'.format(length)
        setattr(self, column, fmt.format(value))

    return property(getter, setter)


def time_property(column, fmt):
    """
    Establish a time property for the given column.
    """
    
    def getter(self):
        value = getattr(self, column)
        if value is None:
            return None
        return datetime.datetime.strptime(value, fmt).time()

    def setter(self, value):
        if value is None:
            setattr(self, column, None)
        else:
            setattr(self, column, value.strftime(fmt))

    return property(getter, setter)


class Department(Base):
    """
    Represents a department.
    """
    __tablename__ = u'Department'

    Department = sa.Column(
        u'Department', sa.String(length=4), primary_key=True, nullable=False)

    Description = sa.Column(
        u'Description', sa.String(length=15), nullable=False)

    FoodStamp = sa.Column(
        u'FoodStamp', sa.String(length=1), nullable=False)

    Tax1 = sa.Column(
        u'Tax1', sa.String(length=1), nullable=False)

    Tax2 = sa.Column(
        u'Tax2', sa.String(length=1), nullable=False)

    Tax3 = sa.Column(
        u'Tax3', sa.String(length=1), nullable=False)

    Discount1 = sa.Column(
        u'Discount1', sa.String(length=1), nullable=False)

    Discount2 = sa.Column(
        u'Discount2', sa.String(length=1), nullable=False)

    FrequentShopper = sa.Column(
        u'FrequentShopper', sa.String(length=1), nullable=False)

    HALO = sa.Column(
        u'HALO', sa.String(length=6), nullable=False)

    halo = decimal_property('HALO', 6)

    LALO = sa.Column(
        u'LALO', sa.String(length=6), nullable=False)

    lalo = decimal_property('LALO', 6)

    QuantityLimit = sa.Column(
        u'QuantityLimit', sa.String(length=2), nullable=False)

    GroupLink = sa.Column(
        u'GroupLink', sa.String(length=2), nullable=False)

    ReceiptCouponNum = sa.Column(
        u'ReceiptCouponNum', sa.String(length=4), nullable=False)

    CouponUnderOver = sa.Column(
        u'CouponUnderOver', sa.String(length=1), nullable=False)

    UnderOverAmount = sa.Column(
        u'UnderOverAmount', sa.String(length=6), nullable=False)

    DepartmentType = sa.Column(
        u'DepartmentType', sa.String(length=1), nullable=False)

    GLAccountNumber = sa.Column(
        u'GLAccountNumber', sa.String(length=10), nullable=False)

    ExcludeFromMinPur = sa.Column(
        u'ExcludeFromMinPur', sa.String(length=1), nullable=False)

    Rx = sa.Column(
        u'Rx', sa.String(length=1), nullable=False)

    NonRxHC = sa.Column(
        u'NonRxHC', sa.String(length=1), nullable=False)

    Blanks = sa.Column(
        u'Blanks', sa.String(length=8), nullable=True)


class Item(Base):
    """
    Represents an item (product).
    """
    __tablename__ = u'Item'

    UPC = sa.Column(
        u'UPC', sa.String(length=20), primary_key=True, nullable=False)

    POSDescription = sa.Column(
        u'POSDescription', sa.String(length=16), nullable=False)

    Department = sa.Column(
        u'Department', sa.String(length=4), nullable=False)

    SubDepartment = sa.Column(
        u'SubDepartment', sa.String(length=4), nullable=False)

    ItemType = sa.Column(
        u'ItemType', sa.String(length=1), nullable=False)

    item_type = int_property('ItemType', 2)

    UnitPrice = sa.Column(
        u'UnitPrice', sa.String(length=6), nullable=False)

    SplitQuantity = sa.Column(
        u'SplitQuantity', sa.String(length=2), nullable=False)

    SplitPrice = sa.Column(
        u'SplitPrice', sa.String(length=6), nullable=False)

    FoodStamp = sa.Column(
        u'FoodStamp', sa.String(length=1), nullable=False)

    Tax1 = sa.Column(
        u'Tax1', sa.String(length=1), nullable=False)

    Tax2 = sa.Column(
        u'Tax2', sa.String(length=1), nullable=False)

    Tax3 = sa.Column(
        u'Tax3', sa.String(length=1), nullable=False)

    Discount1 = sa.Column(
        u'Discount1', sa.String(length=1), nullable=False)

    Discount2 = sa.Column(
        u'Discount2', sa.String(length=1), nullable=False)

    FrequentShopper = sa.Column(
        u'FrequentShopper', sa.String(length=1), nullable=False)

    EnforceQuantity = sa.Column(
        u'EnforceQuantity', sa.String(length=1), nullable=False)

    LimitedQuantity = sa.Column(
        u'LimitedQuantity', sa.String(length=2), nullable=False)

    LimitedPrice = sa.Column(
        u'LimitedPrice', sa.String(length=6), nullable=False)

    BottleLink = sa.Column(
        u'BottleLink', sa.String(length=2), nullable=False)

    bottle_link = int_property('BottleLink', 2)

    MixMatch = sa.Column(
        u'MixMatch', sa.String(length=4), nullable=False)

    ReportCode = sa.Column(
        u'ReportCode', sa.String(length=4), nullable=False)

    AdLevel = sa.Column(
        u'AdLevel', sa.String(length=1), nullable=False)

    VisualVerify = sa.Column(
        u'VisualVerify', sa.String(length=1), nullable=False)

    ReceiptCouponNumber = sa.Column(
        u'ReceiptCouponNumber', sa.String(length=4), nullable=False)

    RestrictSale = sa.Column(
        u'RestrictSale', sa.String(length=1), nullable=False)

    UnitCost = sa.Column(
        u'UnitCost', sa.String(length=9), nullable=False)

    CaseCost = sa.Column(
        u'CaseCost', sa.String(length=9), nullable=False)

    CaseQuantity = sa.Column(
        u'CaseQuantity', sa.String(length=4), nullable=False)

    CommodityCode = sa.Column(
        u'CommodityCode', sa.String(length=8), nullable=False)

    DateTimeModified = sa.Column(
        u'DateTimeModified', sa.String(length=14), nullable=False)

    WIC = sa.Column(
        u'WIC', sa.String(length=1), nullable=False)

    ElectronicCoupon = sa.Column(
        u'ElectronicCoupon', sa.String(length=1), nullable=False)

    AuditFlag = sa.Column(
        u'AuditFlag', sa.String(length=1), nullable=False)

    VendorNumber = sa.Column(
        u'VendorNumber', sa.String(length=8), nullable=False)

    ItemNumber = sa.Column(
        u'ItemNumber', sa.String(length=13), nullable=False)

    ModifiedByBatch = sa.Column(
        u'ModifiedByBatch', sa.String(length=4), nullable=False)

    Seasonal = sa.Column(
        u'Seasonal', sa.String(length=1), nullable=False)

    AdFlag = sa.Column(
        u'AdFlag', sa.String(length=1), nullable=False)

    PackSize = sa.Column(
        u'PackSize', sa.String(length=4), nullable=False)

    UnitSize = sa.Column(
        u'UnitSize', sa.String(length=9), nullable=False)

    UnitMeasure = sa.Column(
        u'UnitMeasure', sa.String(length=4), nullable=False)

    Aisle = sa.Column(
        u'Aisle', sa.String(length=3), nullable=False)

    Shelf = sa.Column(
        u'Shelf', sa.String(length=3), nullable=False)

    Location = sa.Column(
        u'Location', sa.String(length=3), nullable=False)

    TagStockNumber = sa.Column(
        u'TagStockNumber', sa.String(length=3), nullable=False)

    TagQuantity = sa.Column(
        u'TagQuantity', sa.String(length=3), nullable=False)

    FamilyCode1 = sa.Column(
        u'FamilyCode1', sa.String(length=3), nullable=False)

    FamilyCode2 = sa.Column(
        u'FamilyCode2', sa.String(length=3), nullable=False)

    TareCode = sa.Column(
        u'TareCode', sa.String(length=2), nullable=False)

    Weight = sa.Column(
        u'Weight', sa.String(length=6), nullable=False)

    AdjectivePriceCode = sa.Column(
        u'AdjectivePriceCode', sa.String(length=4), nullable=False)

    adjective_price = relationship('AdjectivePrice',
                                   primaryjoin='AdjectivePrice.AdjectivePriceCode == foreign(Item.AdjectivePriceCode)',
                                   uselist=False,
                                   doc="""
Reference to the :class:`AdjectivePrice` with which the product is associated,
if any.
""")

    Points = sa.Column(
        u'Points', sa.String(length=6), nullable=False)

    BonusPoints = sa.Column(
        u'BonusPoints', sa.String(length=6), nullable=False)

    FreeItem = sa.Column(
        u'FreeItem', sa.String(length=1), nullable=False)

    ExcludeFromMinPur = sa.Column(
        u'ExcludeFromMinPur', sa.String(length=1), nullable=False)

    BOFDescription = sa.Column(
        u'BOFDescription', sa.String(length=30), nullable=False)

    GiftCard = sa.Column(
        u'GiftCard', sa.String(length=1), nullable=False)

    PriceOverride = sa.Column(
        u'PriceOverride', sa.String(length=1), nullable=False)

    LikeCode = sa.Column(
        u'LikeCode', sa.String(length=20), nullable=False)

    PriceEndDate = sa.Column(
        u'PriceEndDate', sa.String(length=8), nullable=False)

    UnitSizeDescription = sa.Column(
        u'UnitSizeDescription', sa.String(length=9), nullable=False)

    BaseUnitPrice = sa.Column(
        u'BaseUnitPrice', sa.String(length=6), nullable=False)

    BaseSplitQuantity = sa.Column(
        u'BaseSplitQuantity', sa.String(length=2), nullable=False)

    BaseSplitPrice = sa.Column(
        u'BaseSplitPrice', sa.String(length=6), nullable=False)

    BaseTareCode = sa.Column(
        u'BaseTareCode', sa.String(length=2), nullable=False)

    BaseWeight = sa.Column(
        u'BaseWeight', sa.String(length=6), nullable=False)

    BaseLimitQuantity = sa.Column(
        u'BaseLimitQuantity', sa.String(length=2), nullable=False)

    BaseLimitPrice = sa.Column(
        u'BaseLimitPrice', sa.String(length=6), nullable=False)

    BaseLikeCode = sa.Column(
        u'BaseLikeCode', sa.String(length=20), nullable=False)

    BaseBottleLink = sa.Column(
        u'BaseBottleLink', sa.String(length=2), nullable=False)

    BaseMixMatch = sa.Column(
        u'BaseMixMatch', sa.String(length=4), nullable=False)

    BaseItemType = sa.Column(
        u'BaseItemType', sa.String(length=1), nullable=False)

    AllowManualWeight = sa.Column(
        u'AllowManualWeight', sa.String(length=1), nullable=False)

    PromoID = sa.Column(
        u'PromoID', sa.String(length=20), nullable=False)

    UOMCompareID = sa.Column(
        u'UOMCompareID', sa.String(length=4), nullable=False)

    Rx = sa.Column(
        u'Rx', sa.String(length=1), nullable=False)

    NonRxHC = sa.Column(
        u'NonRxHC', sa.String(length=1), nullable=False)

    CVV = sa.Column(
        u'CVV', sa.String(length=1), nullable=False)

    Blanks = sa.Column(
        u'Blanks', sa.String(length=173), nullable=True)

    def __repr__(self):
        return u'Item(UPC={0})'.format(repr(self.UPC))

    def __unicode__(self):
        return unicode(self.POSDescription)


class AdjectivePrice(Base):
    """
    Represents an adjective pricing structure.
    """
    __tablename__ = 'AdjectivePrice'

    AdjectivePriceCode = sa.Column(
        'AdjectivePriceCode', sa.String(length=4), primary_key=True, nullable=False)

    AdjectivePriceType = sa.Column(
        'AdjectivePriceType', sa.String(length=1), nullable=False)

    DESCRIPTION_1 = sa.Column(
        'DESCRIPTION_1', sa.String(length=16), nullable=False)

    DESCRIPTION_2 = sa.Column(
        'DESCRIPTION_2', sa.String(length=16), nullable=False)

    DESCRIPTION_3 = sa.Column(
        'DESCRIPTION_3', sa.String(length=16), nullable=False)

    DESCRIPTION_4 = sa.Column(
        'DESCRIPTION_4', sa.String(length=16), nullable=False)

    DESCRIPTION_5 = sa.Column(
        'DESCRIPTION_5', sa.String(length=16), nullable=False)

    QUANTITY_2 = sa.Column(
        'QUANTITY_2', sa.String(length=4), nullable=False)

    quantity_2 = int_property('QUANTITY_2', 4)

    QUANTITY_3 = sa.Column(
        'QUANTITY_3', sa.String(length=4), nullable=False)

    QUANTITY_4 = sa.Column(
        'QUANTITY_4', sa.String(length=4), nullable=False)

    QUANTITY_5 = sa.Column(
        'QUANTITY_5', sa.String(length=4), nullable=False)

    PRICE_2 = sa.Column(
        'PRICE_2', sa.String(length=6), nullable=False)

    price_2 = decimal_property('PRICE_2', 6)

    PRICE_3 = sa.Column(
        'PRICE_3', sa.String(length=6), nullable=False)

    PRICE_4 = sa.Column(
        'PRICE_4', sa.String(length=6), nullable=False)

    PRICE_5 = sa.Column(
        'PRICE_5', sa.String(length=6), nullable=False)

    GENERALDESCRIPTION = sa.Column(
        'GENERALDESCRIPTION', sa.String(length=20), nullable=False)

    BLANKS = sa.Column(
        'BLANKS', sa.String(length=16), nullable=True)


class ElectronicCoupon(Base):
    """
    Represents an electronic coupon.
    """
    __tablename__ = u'ElectronicCoupon'

    Upc = sa.Column(
        u'Upc', sa.String(length=20), primary_key=True, nullable=False)

    Vendor = sa.Column(
        u'Vendor', sa.String(length=12), nullable=False)

    Commoditycode = sa.Column(
        u'Commoditycode', sa.String(length=12), nullable=False)

    Department = sa.Column(
        u'Department', sa.String(length=4), primary_key=True, nullable=False)

    Reportcode = sa.Column(
        u'Reportcode', sa.String(length=4), primary_key=True, nullable=False)

    Description = sa.Column(
        u'Description', sa.String(length=30), nullable=False)

    Coupontype = sa.Column(
        u'Coupontype', sa.String(length=1), nullable=False)

    Discounttype = sa.Column(
        u'Discounttype', sa.String(length=1), nullable=False)

    Discount = sa.Column(
        u'Discount', sa.String(length=6), nullable=False)

    discount = decimal_property('Discount', 6)

    Buys = sa.Column(
        u'Buys', sa.String(length=6), nullable=False)

    Gets = sa.Column(
        u'Gets', sa.String(length=6), nullable=False)

    Limitedquantity = sa.Column(
        u'Limitedquantity', sa.String(length=6), nullable=False)

    Startdate = sa.Column(
        u'Startdate', sa.String(length=6), nullable=False)

    start_date = date_property('Startdate', '%m%d%y')

    Starttime = sa.Column(
        u'Starttime', sa.String(length=4), nullable=False)

    start_time = time_property('Starttime', '%H%M')

    @property
    def starts(self):
        """
        The date and time when the coupon starts.
        """
        if self.start_date is None:
            return self.start_time
        if self.start_time is None:
            return self.start_date
        return datetime.datetime.combine(self.start_date, self.start_time)

    Enddate = sa.Column(
        u'Enddate', sa.String(length=6), nullable=False)

    end_date = date_property('Enddate', '%m%d%y')

    Endtime = sa.Column(
        u'Endtime', sa.String(length=4), nullable=False)

    end_time = time_property('Endtime', '%H%M')

    @property
    def ends(self):
        """
        The date and time when the coupon ends.
        """
        if self.end_date is None:
            return self.end_time
        if self.end_time is None:
            return self.end_date
        return datetime.datetime.combine(self.end_date, self.end_time)

    Points = sa.Column(
        u'Points', sa.String(length=6), nullable=False)

    Bonus = sa.Column(
        u'Bonus', sa.String(length=6), nullable=False)

    ReceiptCoupon = sa.Column(
        u'ReceiptCoupon', sa.String(length=4), nullable=False)

    PromoID = sa.Column(
        u'PromoID', sa.String(length=20), primary_key=True, nullable=False)

    CEcquantity = sa.Column(
        u'CEcquantity', sa.String(length=6), nullable=False)

    CEcamount = sa.Column(
        u'CEcamount', sa.String(length=9), nullable=False)

    CTpoints = sa.Column(
        u'CTpoints', sa.String(length=12), nullable=False)

    CTbonus = sa.Column(
        u'CTbonus', sa.String(length=12), nullable=False)

    PECquantity = sa.Column(
        u'PECquantity', sa.String(length=6), nullable=False)

    PECcamount = sa.Column(
        u'PECcamount', sa.String(length=9), nullable=False)

    PTpoints = sa.Column(
        u'PTpoints', sa.String(length=12), nullable=False)

    PTbonus = sa.Column(
        u'PTbonus', sa.String(length=12), nullable=False)

    CloseDate = sa.Column(
        u'CloseDate', sa.String(length=6), nullable=False)

    FrequentShopperLevel = sa.Column(
        u'FrequentShopperLevel', sa.String(length=1), primary_key=True, nullable=False)

    frequent_shopper_level = int_property('FrequentShopperLevel', 1)

    MinPurchase = sa.Column(
        u'MinPurchase', sa.String(length=7), nullable=False)

    MixMatch = sa.Column(
        u'MixMatch', sa.String(length=4), primary_key=True, nullable=False)

    TotalSale = sa.Column(
        u'TotalSale', sa.String(length=7), primary_key=True, nullable=False)

    DepartmentTotal = sa.Column(
        u'DepartmentTotal', sa.String(length=7), nullable=False)

    SortDepartment = sa.Column(
        u'SortDepartment', sa.String(length=4), nullable=False)

    RedeemValue = sa.Column(
        u'RedeemValue', sa.String(length=6), nullable=False)

    BundleCode = sa.Column(
        u'BundleCode', sa.String(length=2), primary_key=True, nullable=False)

    Weighted = sa.Column(
        u'Weighted', sa.String(length=1), nullable=False)

    FavorFlag = sa.Column(
        u'FavorFlag', sa.String(length=1), nullable=False)

    Mediafile = sa.Column(
        u'Mediafile', sa.String(length=35), nullable=False)

    UCIPromoID = sa.Column(
        u'UCIPromoID', sa.String(length=20), nullable=False)

    Unused = sa.Column(
        u'Unused', sa.String(length=10), nullable=False)

    MMCurrentQuantity = sa.Column(
        u'MMCurrentQuantity', sa.String(length=6), nullable=False)

    MMPTDQuantity = sa.Column(
        u'MMPTDQuantity', sa.String(length=6), nullable=False)

    MinimumQty = sa.Column(
        u'MinimumQty', sa.String(length=6), nullable=False)

    Blanks = sa.Column(
        u'Blanks', sa.String(length=98), nullable=True)


class Customer(Base):
    """
    Represents a customer.
    """

    __tablename__ = 'Customer'

    account_number = sa.Column(
        'AccountNumber', sa.String(length=12), primary_key=True, nullable=False)

    last_name = sa.Column(
        'LastName', sa.String(length=25), nullable=False, default='')

    first_name = sa.Column(
        'FirstName', sa.String(length=25), nullable=False, default='')

    address = sa.Column(
        'Address', sa.String(length=25), nullable=False, default='')

    city = sa.Column(
        'City', sa.String(length=15), nullable=False, default='')

    state = sa.Column(
        'State', sa.String(length=2), nullable=False, default='')

    zipcode = sa.Column(
        'Zipcode', sa.String(length=9), nullable=False, default='')

    phone_number = sa.Column(
        'PhoneNumber', sa.String(length=10), nullable=False, default='0000000000')

    social_security_number = sa.Column(
        'SSNumber', sa.String(length=9), nullable=False, default='000000000')

    _DateOpened = sa.Column(
        'DateOpened', sa.String(length=6), nullable=False, default='000000')

    date_opened = date_property('_DateOpened', '%m%d%y')

    tax_exempt_code = sa.Column(
        'TaxExemptCode', sa.String(length=12), nullable=False, default='000000000000')

    checking_account_number = sa.Column(
        'CheckingAcctNum', sa.String(length=12), nullable=False, default='000000000000')

    _NumVisitsTD = sa.Column(
        'NumVisitsTD', sa.String(length=3), nullable=False, default='000')

    number_visits_to_date = int_property('_NumVisitsTD', 3)

    _NumVisitsPTD = sa.Column(
        'NumVisitsPTD', sa.String(length=3), nullable=False, default='000')

    number_visits_period_to_date = int_property('_NumVisitsPTD', 3)

    purchase_amount_to_date = sa.Column(
        'PurchaseAmountTD', sa.String(length=9), nullable=False, default='000000000')

    purchase_amount_period_to_date = sa.Column(
        'PurchaseAmountPTD', sa.String(length=9), nullable=False, default='000000000')

    _DOLPurchase = sa.Column(
        'DOLPurchase', sa.String(length=6), nullable=False, default='000000')

    date_of_last_purchase = date_property('_DOLPurchase', '%m%d%y')

    comments = sa.Column(
        'Comments', sa.String(length=30), nullable=False, default='')

    override_count = sa.Column(
        'OverrideCount', sa.String(length=3), nullable=False, default='000')

    account_status = sa.Column(
        'AccountStatus', sa.String(length=1), nullable=False, default='0')

    _FSDollarsTD = sa.Column(
        'FSDollarsTD', sa.String(length=9), nullable=False, default='000000000')

    frequent_shopper_dollars_to_date = decimal_property('_FSDollarsTD', 9)

    _FSDollarsPTD = sa.Column(
        'FSDollarsPTD', sa.String(length=9), nullable=False, default='000000000')

    frequent_shopper_dollars_period_to_date = decimal_property('_FSDollarsPTD', 9)

    number_checks_today = sa.Column(
        'NumChecksToday', sa.String(length=1), nullable=False, default='0')

    number_checks_this_week = sa.Column(
        'NumChecksThisWeek', sa.String(length=1), nullable=False, default='0')

    amount_checks_today = sa.Column(
        'AmtChecksToday', sa.String(length=9), nullable=False, default='000000000')

    amount_checks_this_week = sa.Column(
        'AmtChecksThisWeek', sa.String(length=9), nullable=False, default='000000000')

    record_changed = sa.Column(
        'RecordChanged', sa.String(length=1), nullable=False, default='0')

    update_flag = sa.Column(
        'UpdateFlag', sa.String(length=1), nullable=False, default='0')

    # TODO: Give this column a default value?  Might need to be business-specific?
    frequent_shopper_level = sa.Column(
        'FSLevel', sa.String(length=1), nullable=False)

    points_period_to_date = sa.Column(
        'PointsPTD', sa.String(length=9), nullable=False, default='000000000')

    points_to_date = sa.Column(
        'PointsTD', sa.String(length=9), nullable=False, default='000000000')

    bonus_points_period_to_date = sa.Column(
        'BonusPointsPTD', sa.String(length=9), nullable=False, default='000000000')

    bonus_points_to_date = sa.Column(
        'BonusPointsTD', sa.String(length=9), nullable=False, default='000000000')

    special_promotion_points = sa.Column(
        'SpecialPromoPoints', sa.String(length=9), nullable=False, default='000000000')

    frequent_shopper_discounts_to_date = sa.Column(
        'FSDiscountsTD', sa.String(length=9), nullable=False, default='000000000')

    frequent_shopper_discounts_period_to_date = sa.Column(
        'FSDiscountsPTD', sa.String(length=9), nullable=False, default='000000000')

    electronic_coupon_to_date = sa.Column(
        'ElectronicCouponTD', sa.String(length=9), nullable=False, default='000000000')

    electronic_coupon_period_to_date = sa.Column(
        'ElectronicCouponPTD', sa.String(length=9), nullable=False, default='000000000')

    auto_discount1_flag = sa.Column(
        'AutoDisc1Flag', sa.String(length=1), nullable=False, default='0')

    auto_discount2_flag = sa.Column(
        'AutoDisc2Flag', sa.String(length=1), nullable=False, default='0')

    auto_discount4_flag = sa.Column(
        'AutoDisc4Flag', sa.String(length=1), nullable=False, default='0')

    auto_discount5_flag = sa.Column(
        'AutoDisc5Flag', sa.String(length=1), nullable=False, default='0')

    cash_only = sa.Column(
        'CashOnly', sa.String(length=1), nullable=False, default='0')

    check_account_number = sa.Column(
        'ChkAcctNum', sa.String(length=20), nullable=False, default='00000000000000000000')

    transit_number = sa.Column(
        'TransitNum', sa.String(length=9), nullable=False, default='000000000')

    blanks = sa.Column(
        'Blanks', sa.String(length=12), nullable=True, default='000000000000')

    def __repr__(self):
        return "Customer(account_number={0})".format(
            repr(self.account_number))


class CustomerExtended(Base):
    """
    Represents extended information about a customer.
    """

    __tablename__ = 'CustomerExtended'

    account_number = sa.Column(
        'AccountNumber', sa.String(length=12), primary_key=True, nullable=True)

    address2 = sa.Column(
        'Address2', sa.String(length=25), nullable=True, default='')

    email = sa.Column(
        'Email1', sa.String(length=50), nullable=True, default='')

    email2 = sa.Column(
        'Email2', sa.String(length=50), nullable=True, default='')

    ship_to_address = sa.Column(
        'ShipTo1', sa.String(length=25), nullable=True, default='')

    ship_to_address2 = sa.Column(
        'ShipTo2', sa.String(length=25), nullable=True, default='')

    ship_to_city = sa.Column(
        'City', sa.String(length=15), nullable=True, default='')

    ship_to_state = sa.Column(
        'State', sa.String(length=2), nullable=True, default='')

    ship_to_zipcode = sa.Column(
        'Zip', sa.String(length=9), nullable=True, default='')

    contact_last_name = sa.Column(
        'ContactLastName', sa.String(length=25), nullable=True, default='')

    contact_first_name = sa.Column(
        'ContactFirstName', sa.String(length=25), nullable=True, default='')

    phone_number2 = sa.Column(
        'Phone2', sa.String(length=10), nullable=True, default='')

    limit_checks_today = sa.Column(
        'LimitChecksToday', sa.String(length=1), nullable=True, default='0')

    limit_checks_this_week = sa.Column(
        'LimitChecksWeek', sa.String(length=1), nullable=True, default='0')

    limit_dollars_today = sa.Column(
        'LimitDollarDay', sa.String(length=9), nullable=True, default='000000000')

    limit_dollars_this_week = sa.Column(
        'LimitDollarWeek', sa.String(length=9), nullable=True, default='000000000')

    memo = sa.Column(
        'Memo', sa.String(length=100), nullable=True, default='')

    drivers_license_number = sa.Column(
        'DriversLicNum', sa.String(length=18), nullable=True, default='')

    drivers_license_state = sa.Column(
        'DriversLicState', sa.String(length=2), nullable=True, default='')

    alternate_lookup = sa.Column(
        'AltLookUp', sa.String(length=20), nullable=True, default='00000000000000000000')

    blanks = sa.Column(
        'Blanks', sa.String(length=58), nullable=True, default='0000000000000000000000000000000000000000000000000000000000')

    def __repr__(self):
        return "CustomerExtended(account_number={0})".format(
            repr(self.account_number))


Customer.extended = relationship(
    CustomerExtended,
    primaryjoin=CustomerExtended.account_number == Customer.account_number,
    foreign_keys=[CustomerExtended.account_number],
    uselist=False,
    backref='customer')

def getset_factory(collection_class, proxy):
    def getter(obj):
        if obj is None:
            return ''
        return getattr(obj, proxy.value_attr)
    def setter(obj, val):
        setattr(obj, proxy.value_attr, val)
    return getter, setter

def extended_proxy(attr):
    def creator(val):
        obj = CustomerExtended()
        setattr(obj, attr, val)
        return obj
    proxy = association_proxy(
        'extended', attr, creator=creator,
        getset_factory=getset_factory)
    setattr(Customer, attr, proxy)

extended_proxy('address2')
extended_proxy('email')
extended_proxy('email2')
extended_proxy('ship_to_address')
extended_proxy('ship_to_address2')
extended_proxy('ship_to_city')
extended_proxy('ship_to_state')
extended_proxy('ship_to_zipcode')
extended_proxy('contact_last_name')
extended_proxy('contact_first_name')
extended_proxy('phone_number2')
extended_proxy('limit_checks_today')
extended_proxy('limit_checks_this_week')
extended_proxy('limit_dollars_today')
extended_proxy('limit_dollars_this_week')
extended_proxy('memo')
extended_proxy('drivers_license_number')
extended_proxy('drivers_license_state')
extended_proxy('alternate_lookup')


class NegativeCheck(Base):
    """
    Represents a "negative check" record.
    """

    __tablename__ = 'NegativeCheck'

    account_number = sa.Column(
        'AccountNumber', sa.String(length=20), primary_key=True, nullable=False)

    last_name = sa.Column(
        'LastName', sa.String(length=25), nullable=False)

    first_name = sa.Column(
        'FirstName', sa.String(length=25), nullable=False)

    address = sa.Column(
        'Address', sa.String(length=25), nullable=False)

    city = sa.Column(
        'City', sa.String(length=15), nullable=False)

    state = sa.Column(
        'State', sa.String(length=2), nullable=False)

    zipcode = sa.Column(
        'Zipcode', sa.String(length=9), nullable=False)

    phone_number = sa.Column(
        'PhoneNumber', sa.String(length=10), nullable=False)

    social_security_number = sa.Column(
        'SSNumber', sa.String(length=9), nullable=False)

    _DateOpened = sa.Column(
        'DateOpened', sa.String(length=6), nullable=False)

    date_opened = date_property('_DateOpened', '%m%d%y')

    tax_exempt_code = sa.Column(
        'TaxExemptCode', sa.String(length=12), nullable=False)

    checking_account_number = sa.Column(
        'CheckingAcctNum', sa.String(length=20), nullable=False)

    _NumChecksTendered = sa.Column(
        'NumChecksTendered', sa.String(length=3), nullable=False)

    checks_tendered_count = int_property('_NumChecksTendered', 3)

    _NumNSFChecks = sa.Column(
        'NumNSFChecks', sa.String(length=3), nullable=False)

    checks_nsf_count = int_property('_NumNSFChecks', 3)

    _ChecksTenderedAmount = sa.Column(
        'ChecksTenderedAmount', sa.String(length=7), nullable=False)

    checks_tendered_amount = decimal_property('_ChecksTenderedAmount', 7)

    _NSFChecksAmount = sa.Column(
        'NSFChecksAmount', sa.String(length=7), nullable=False)

    checks_nsf_amount = decimal_property('_NSFChecksAmount', 7)

    _DateOfLastPurchase = sa.Column(
        'DateOfLastPurchase', sa.String(length=6), nullable=False)

    date_of_last_purchase = date_property('_DateOfLastPurchase', '%m%d%y')

    comments = sa.Column(
        'Comments', sa.String(length=30), nullable=False)

    _OverrideCount = sa.Column(
        'OverrideCount', sa.String(length=3), nullable=False)

    override_count = int_property('_OverrideCount', 3)

    account_status = sa.Column(
        'AccountStatus', sa.String(length=1), nullable=False)

    transit_routing_number = sa.Column(
        'TransitRoutingNumber', sa.String(length=9), nullable=False)

    blanks = sa.Column(
        'Blanks', sa.String(length=66), nullable=True)

    def __repr__(self):
        return "NegativeCheck(account_number={0})".format(
            repr(self.account_number))


NegativeCheck.customer = relationship(
    Customer,
    primaryjoin=Customer.account_number == NegativeCheck.account_number,
    foreign_keys=[Customer.account_number],
    uselist=False)


class EJTransaction(Base):
    """
    Represents an electronic journal transaction.
    """

    __tablename__ = 'tabEJTransaction'

    transaction_id = sa.Column(
        'TransactionID', sa.Integer(), primary_key=True, nullable=False)

    transaction_number = sa.Column(
        'TransactionNum', sa.String(length=8), nullable=False)

    cashier_id = sa.Column(
        'CashierID', sa.SmallInteger(), nullable=False)

    cashier_name = sa.Column(
        'CashierName', sa.String(length=20), nullable=True)

    register_number = sa.Column(
        'RegisterNum', sa.SmallInteger(), nullable=False)

    start_date = sa.Column(
        'StartDate', sa.Date(), nullable=False)

    end_date = sa.Column(
        'EndDate', sa.Date(), nullable=False)

    start_time = sa.Column(
        'StartTime', sa.Time(), nullable=False)

    end_time = sa.Column(
        'EndTime', sa.Time(), nullable=False)

    @property
    def started(self):
        """
        The date and time when the transaction started.
        """
        if self.start_date is None:
            return self.start_time
        if self.start_time is None:
            return self.start_date
        return datetime.datetime.combine(self.start_date, self.start_time)

    @property
    def ended(self):
        """
        The date and time when the transaction ended.
        """
        if self.end_date is None:
            return self.end_time
        if self.end_time is None:
            return self.end_date
        return datetime.datetime.combine(self.end_date, self.end_time)

    transaction_type = sa.Column(
        'TransactionType', sa.SmallInteger(), nullable=False)

    shoppers_hotline = sa.Column(
        'ShoppersHotline', sa.String(length=12), nullable=True)

    transaction_amount = sa.Column(
        'TransactionAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    item_subtotal_amount = sa.Column(
        'ItemSubtotalAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    total_discount_amount = sa.Column(
        'TotalDiscAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    total_void_amount = sa.Column(
        'TotalVoidAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    total_return_amount = sa.Column(
        'TotalReturnAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    total_tt_discount_amount = sa.Column(
        'TotalTTDiscAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    number_of_items = sa.Column(
        'NumberOfItems', sa.SmallInteger(), nullable=True)

    original_transaction_number = sa.Column(
        'OrigTransactionNum', sa.String(length=8), nullable=True)

    retrieved_flag = sa.Column(
        'RetrievedFlag', sa.Boolean(), nullable=False)

    pf_recovery_flag = sa.Column(
        'PFRecoveryFlag', sa.Boolean(), nullable=False)

    cancel_flag = sa.Column(
        'CancelFlag', sa.Boolean(), nullable=False)

    suspend_flag = sa.Column(
        'SuspendFlag', sa.Boolean(), nullable=False)

    online_flag = sa.Column(
        'OnlineFlag', sa.SmallInteger(), nullable=True)

    register_version = sa.Column(
        'RegisterVer', sa.String(length=15), nullable=True)

    transaction_seconds = sa.Column(
        'TransSeconds', sa.Integer(), nullable=True)

    idle_seconds = sa.Column(
        'IdleSeconds', sa.Integer(), nullable=True)

    ring_seconds = sa.Column(
        'RingSeconds', sa.Integer(), nullable=True)

    tender_seconds = sa.Column(
        'TenderSeconds', sa.Integer(), nullable=True)

    scan_rate = sa.Column(
        'ScanRate', sa.Numeric(precision=10, scale=2), nullable=True)

    scanned_entries = sa.Column(
        'ScannedEntries', sa.SmallInteger(), nullable=True)

    keyed_entries = sa.Column(
        'KeyedEntries', sa.SmallInteger(), nullable=True)

    open_department_entries = sa.Column(
        'OpenDeptEntries', sa.SmallInteger(), nullable=True)

    aff_flag = sa.Column(
        'AFFFlag', sa.SmallInteger(), nullable=True)

    def __repr__(self):
        return "EJTransaction(transaction_id={0}, transaction_number={1})".format(
            repr(self.transaction_id), repr(self.transaction_number))


class EJCustomer(Base):
    """
    Represents a customer association with an EJ transaction.
    """

    __tablename__ = 'tabEJCustomer'

    transaction_id = sa.Column(
        'TransactionID', sa.Integer(), primary_key=True, nullable=False)

    customer_number = sa.Column(
        'CustomerNum', sa.String(length=12), primary_key=True, nullable=False)

    last_name = sa.Column(
        'LastName', sa.String(length=25), nullable=True)

    first_name = sa.Column(
        'FirstName', sa.String(length=25), nullable=True)

    restricted_action_flag = sa.Column(
        'RestrictedActionFlag', sa.SmallInteger(), nullable=True)

    approval_mode = sa.Column(
        'ApprovalMode', sa.SmallInteger(), nullable=True)

    birthdate = sa.Column(
        'Birthdate', sa.Date(), nullable=True)

    verification_id = sa.Column(
        'VerificationID', sa.String(length=20), nullable=True)

    origin = sa.Column(
        'Origin', sa.SmallInteger(), nullable=True)

    frequent_shopper_level = sa.Column(
        'FreqShopLevel', sa.SmallInteger(), nullable=True)

    def __repr__(self):
        return "EJCustomer(transaction_id={0}, customer_number={1})".format(
            repr(self.transaction_id), repr(self.customer_number))


EJCustomer.customer = relationship(
    Customer,
    primaryjoin=Customer.account_number == EJCustomer.customer_number,
    foreign_keys=[Customer.account_number],
    uselist=False)

EJTransaction.customers = relationship(
    EJCustomer,
    primaryjoin=EJCustomer.transaction_id == EJTransaction.transaction_id,
    foreign_keys=[EJCustomer.transaction_id],
    backref='transaction')


class EJItem(Base):
    """
    Represents a line item within an EJ transaction.
    """

    __tablename__ = 'tabEJItem'

    transaction_id = sa.Column(
        'TransactionID', sa.Integer(), primary_key=True, nullable=False)

    sequence_number = sa.Column(
        'SequenceNum', sa.Integer(), primary_key=True, nullable=False)

    item_type = sa.Column(
        'ItemType', sa.SmallInteger(), nullable=False)

    upc = sa.Column(
        'UPC', sa.String(length=20), nullable=False)

    item_description = sa.Column(
        'ItemDesc', sa.String(length=20), nullable=True)

    department_number = sa.Column(
        'DepartmentNum', sa.SmallInteger(), nullable=False)

    subdepartment_number = sa.Column(
        'SubDeptNum', sa.SmallInteger(), nullable=True)

    quantity = sa.Column(
        'Quantity', sa.Integer(), nullable=False)

    adjective_quantity = sa.Column(
        'AdjectiveQty', sa.Integer(), nullable=True)

    adjective_level = sa.Column(
        'AdjectiveLevel', sa.SmallInteger(), nullable=True)

    sold_at_price = sa.Column(
        'SoldAtPrice', sa.Numeric(precision=19, scale=4), nullable=True)

    discount_amount = sa.Column(
        'DiscountAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    bottle_deposit_amount = sa.Column(
        'BottleDepAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    original_quantity = sa.Column(
        'OriginalQty', sa.Integer(), nullable=True)

    original_price = sa.Column(
        'OriginalPrice', sa.Numeric(precision=19, scale=4), nullable=True)

    regular_price = sa.Column(
        'RegularPrice', sa.Numeric(precision=19, scale=4), nullable=True)

    food_stamp_status = sa.Column(
        'FoodStampStatus', sa.Boolean(), nullable=False)

    tax1_status = sa.Column(
        'Tax1Status', sa.Boolean(), nullable=False)

    tax2_status = sa.Column(
        'Tax2Status', sa.Boolean(), nullable=False)

    tax3_status = sa.Column(
        'Tax3Status', sa.Boolean(), nullable=False)

    discount1_status = sa.Column(
        'Disc1Status', sa.Boolean(), nullable=False)

    discount2_status = sa.Column(
        'Disc2Status', sa.Boolean(), nullable=False)

    discount4_status = sa.Column(
        'Disc4Status', sa.Boolean(), nullable=False)

    discount5_status = sa.Column(
        'Disc5Status', sa.Boolean(), nullable=False)

    frequent_shopper_status = sa.Column(
        'FreqShopperStatus', sa.Boolean(), nullable=False)

    sale_level = sa.Column(
        'SaleLevel', sa.SmallInteger(), nullable=True)

    bottle_link_department_number = sa.Column(
        'BottleLinkDeptNum', sa.SmallInteger(), nullable=True)

    scale_weight = sa.Column(
        'ScaleWeight', sa.Numeric(precision=5, scale=3), nullable=True)

    split_weight = sa.Column(
        'SplitWeight', sa.Numeric(precision=6, scale=3), nullable=True)

    tare_weight = sa.Column(
        'TareWeight', sa.Numeric(precision=5, scale=3), nullable=True)

    tare_code = sa.Column(
        'TareCode', sa.SmallInteger(), nullable=True)

    commodity_code = sa.Column(
        'CommodityCode', sa.String(length=8), nullable=True)

    reason_code = sa.Column(
        'ReasonCode', sa.SmallInteger(), nullable=True)

    report_code = sa.Column(
        'ReportCode', sa.String(length=4), nullable=True)

    online_flag = sa.Column(
        'OnlineFlag', sa.Boolean(), nullable=False)

    void_flag = sa.Column(
        'VoidFlag', sa.Boolean(), nullable=False)

    error_correct_flag = sa.Column(
        'ErrorCorrectFlag', sa.Boolean(), nullable=False)

    scan_flag = sa.Column(
        'ScanFlag', sa.Boolean(), nullable=False)

    override_flag = sa.Column(
        'OverrideFlag', sa.Boolean(), nullable=False)

    return_flag = sa.Column(
        'ReturnFlag', sa.Boolean(), nullable=False)

    refund_flag = sa.Column(
        'RefundFlag', sa.Boolean(), nullable=False)

    not_found_flag = sa.Column(
        'NotFoundFlag', sa.Boolean(), nullable=False)

    suspended_flag = sa.Column(
        'SuspendedFlag', sa.Boolean(), nullable=False)

    retrieved_flag = sa.Column(
        'RetrievedFlag', sa.Boolean(), nullable=False)

    canceled_flag = sa.Column(
        'CanceledFlag', sa.Boolean(), nullable=False)

    unit_price = sa.Column(
        'UnitPrice', sa.Numeric(precision=19, scale=4), nullable=True)

    case_cost = sa.Column(
        'CaseCost', sa.Numeric(precision=19, scale=4), nullable=True)

    case_quantity = sa.Column(
        'CaseQty', sa.Integer(), nullable=True)

    mix_match_code = sa.Column(
        'MixMatchCode', sa.Integer(), nullable=True)

    coupon_code = sa.Column(
        'CouponCode', sa.Integer(), nullable=True)

    fsa_rx_flag = sa.Column(
        'FSARxFlag', sa.Boolean(), nullable=False)

    fsa_non_rx_flag = sa.Column(
        'FSANonRxFlag', sa.Boolean(), nullable=False)

    promotion_id = sa.Column(
        'PromoID', sa.String(length=20), nullable=True)

    def __repr__(self):
        return "EJItem(transaction_id={0}, sequence_number={1})".format(
            repr(self.transaction_id), repr(self.sequence_number))


EJTransaction.items = relationship(
    EJItem,
    primaryjoin=EJItem.transaction_id == EJTransaction.transaction_id,
    foreign_keys=[EJItem.transaction_id],
    order_by=EJItem.sequence_number,
    backref='transaction')


class EJPromotion(Base):
    """
    Represents a promotion within an EJ transaction.
    """

    __tablename__ = 'tabEJPromotions'

    transaction_id = sa.Column(
        'TransactionID', sa.Integer(), primary_key=True, nullable=False)

    sequence_number = sa.Column(
        'SequenceNum', sa.Integer(), primary_key=True, nullable=False)

    promotion_number = sa.Column(
        'PromotionNum', sa.String(length=20), nullable=True)

    promotion_type = sa.Column(
        'PromoType', sa.SmallInteger(), nullable=True)

    quantity = sa.Column(
        'Quantity', sa.Integer(), nullable=True)

    amount = sa.Column(
        'Amount', sa.Numeric(precision=19, scale=4), nullable=True)

    department_number = sa.Column(
        'DepartmentNum', sa.SmallInteger(), nullable=True)

    frequent_shopper_level = sa.Column(
        'FreqShopperLevel', sa.SmallInteger(), nullable=True)

    coupon_trigger = sa.Column(
        'CouponTrigger', sa.SmallInteger(), nullable=True)

    points = sa.Column(
        'Points', sa.Integer(), nullable=True)

    bonus_points = sa.Column(
        'BonusPoints', sa.Integer(), nullable=True)

    promotion_description = sa.Column(
        'PromoDesc', sa.String(length=30), nullable=True)

    multimedia_filename = sa.Column(
        'MultiMediaFN', sa.String(length=35), nullable=True)

    void_flag = sa.Column(
        'VoidFlag', sa.Boolean(), nullable=False)

    error_correct_flag = sa.Column(
        'ErrorCorrectFlag', sa.Boolean(), nullable=False)

    def __repr__(self):
        return "EJPromotion(transaction_id={0}, sequence_number={1})".format(
            repr(self.transaction_id), repr(self.sequence_number))


EJTransaction.promotions = relationship(
    EJPromotion,
    primaryjoin=EJPromotion.transaction_id == EJTransaction.transaction_id,
    foreign_keys=[EJPromotion.transaction_id],
    order_by=EJPromotion.sequence_number,
    backref='transaction')


class EJDiscount(Base):
    """
    Represents a discount on an EJ transaction.
    """

    __tablename__ = 'tabEJDiscount'

    transaction_id = sa.Column(
        'TransactionID', sa.Integer(), primary_key=True, nullable=False)

    sequence_number = sa.Column(
        'SequenceNum', sa.Integer(), primary_key=True, nullable=False)

    discount_number = sa.Column(
        'DiscountNum', sa.SmallInteger(), nullable=False)

    discount_type = sa.Column(
        'DiscountType', sa.SmallInteger(), nullable=True)

    discount_amount = sa.Column(
        'DiscountAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    def __repr__(self):
        return "EJDiscount(transaction_id={0}, sequence_number={1})".format(
            repr(self.transaction_id), repr(self.sequence_number))


EJTransaction.discounts = relationship(
    EJDiscount,
    primaryjoin=EJDiscount.transaction_id == EJTransaction.transaction_id,
    foreign_keys=[EJDiscount.transaction_id],
    order_by=EJDiscount.sequence_number,
    backref='transaction')


class EJTax(Base):
    """
    Represents the tax on an EJ transaction.
    """

    __tablename__ = 'tabEJTax'

    transaction_id = sa.Column(
        'TransactionID', sa.Integer(), primary_key=True, nullable=False)

    tax_number = sa.Column(
        'TaxNum', sa.SmallInteger(), primary_key=True, nullable=False)

    tax_amount = sa.Column(
        'TaxAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    taxable_sales_amount = sa.Column(
        'TaxableSalesAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    food_stamp_taxable_sales_amount = sa.Column(
        'FSTaxableSalesAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    food_stamp_tax_amount = sa.Column(
        'FSTaxAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    tax_exempt_amount = sa.Column(
        'TaxExemptAmt', sa.Numeric(precision=19, scale=4), nullable=True)

    tax_exempt_code = sa.Column(
        'TaxExemptCode', sa.String(length=12), nullable=True)

    def __repr__(self):
        return "EJTax(transaction_id={0}, tax_number={1})".format(
            repr(self.transaction_id), repr(self.tax_number))


EJTransaction.taxes = relationship(
    EJTax,
    primaryjoin=EJTax.transaction_id == EJTransaction.transaction_id,
    foreign_keys=[EJTax.transaction_id],
    order_by=EJTax.tax_number,
    backref='transaction')


class InstoreCharge(Base):
    """
    Represents an in-store charge account.
    """

    __tablename__ = 'InstoreCharge'

    account_number = sa.Column(
        'ACCOUNTNUMBER', sa.String(length=12), primary_key=True, nullable=False)

    _DATEOPENED = sa.Column(
        'DATEOPENED', sa.String(length=6), nullable=False, default='000000')

    date_opened = date_property('_DATEOPENED', '%m%d%y')

    _CREDITLIMIT = sa.Column(
        'CREDITLIMIT', sa.String(length=6), nullable=False, default='000000')

    credit_limit = decimal_property('_CREDITLIMIT', 6)

    credit_status = sa.Column(
        'CREDITSTATUS', sa.String(length=1), nullable=False, default='0')

    _DATEOFLASTPAYMENT = sa.Column(
        'DATEOFLASTPAYMENT', sa.String(length=6), nullable=False, default='000000')

    date_of_last_payment = date_property('_DATEOFLASTPAYMENT', '%y%m%d')

    _AMOUNTOFLASTPAYMENT = sa.Column(
        'AMOUNTOFLASTPAYMENT', sa.String(length=8), nullable=False, default='00000000')

    amount_of_last_payment = decimal_property('_AMOUNTOFLASTPAYMENT', 8)

    _CURRENTBALANCE = sa.Column(
        'CURRENTBALANCE', sa.String(length=8), nullable=False, default='00000000')

    current_balance = decimal_property('_CURRENTBALANCE', 8)

    _PREVIOUSBALANCE = sa.Column(
        'PREVIOUSBALANCE', sa.String(length=8), nullable=False, default='00000000')

    previous_balance = decimal_property('_PREVIOUSBALANCE', 8)

    _PAGENUMBER = sa.Column(
        'PAGENUMBER', sa.String(length=2), nullable=False, default='00')

    page_number = int_property('_PAGENUMBER', 2)

    _LINENUMBER = sa.Column(
        'LINENUMBER', sa.String(length=2), nullable=False, default='00')

    line_number = int_property('_LINENUMBER', 2)

    finance_charges = sa.Column(
        'FINANCECHARGES', sa.String(length=1), nullable=False, default='0')

    apr = sa.Column(
        'APR', sa.String(length=4), nullable=False, default='0000')

    blanks = sa.Column(
        'BLANKS', sa.String(length=34), nullable=True, default='0000000000000000000000000000000000')

    update_flag = sa.Column(
        'UPDATEFLAG', sa.String(length=1), nullable=False, default='0')

    def __repr__(self):
        return "InstoreCharge(account_number={0})".format(
            repr(self.account_number))


InstoreCharge.customer = relationship(
    Customer,
    primaryjoin=Customer.account_number == InstoreCharge.account_number,
    foreign_keys=[Customer.account_number],
    uselist=False)


class ChargeHistory(Base):
    """
    Represents a charge event within a customer account's history.
    """

    __tablename__ = 'ChargeHistory'

    account_number = sa.Column(
        'ACCOUNTNUMBER', sa.String(length=12), primary_key=True, nullable=True)

    transaction_number = sa.Column(
        'TRANSACTIONNUMBER', sa.String(length=8), primary_key=True, nullable=True)

    _BUSINESSDATE = sa.Column(
        'BUSINESSDATE', sa.String(length=6), nullable=True)

    business_date = date_property('_BUSINESSDATE', '%y%m%d')

    _BUSINESSTIME = sa.Column(
        'BUSINESSTIME', sa.String(length=4), nullable=True)

    business_time = time_property('_BUSINESSTIME', '%H%M')

    @property
    def occurred(self):
        """
        The date and time when the charge occurred.
        """
        if self.business_date is None:
            return self.business_time
        if self.business_time is None:
            return self.business_date
        return datetime.datetime.combine(self.business_date, self.business_time)

    store_number = sa.Column(
        'STORENUMBER', sa.String(length=4), nullable=True)

    register_number = sa.Column(
        'REGISTERNUMBER', sa.String(length=2), nullable=True)

    cashier_number = sa.Column(
        'CASHIERNUMBER', sa.String(length=3), nullable=True)

    activity_code = sa.Column(
        'ACTIVITYCODE', sa.String(length=1), nullable=True)

    _AMOUNT = sa.Column(
        'AMOUNT', sa.String(length=8), nullable=True)

    amount = decimal_property('_AMOUNT', 8)

    export_flag = sa.Column(
        'EXPORTFLAG', sa.String(length=1), nullable=True)

    update_flag = sa.Column(
        'UPDATEFLAG', sa.String(length=1), nullable=True)

    def __repr__(self):
        return "ChargeHistory(account_number={0}, transaction_number={1})".format(
            repr(self.account_number), repr(self.transaction_number))


EJTransaction.charges = relationship(
    ChargeHistory,
    primaryjoin=ChargeHistory.transaction_number == EJTransaction.transaction_number,
    foreign_keys=[ChargeHistory.transaction_number],
    backref='transaction')
